#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_matrix.h>
#include <mef.h>

/*
import numpy
from scipy.stats import multivariate_normal
mu    = numpy.array([0., 0.])
Sigma = numpy.array([[1., 0.],[0., 2.]])

data = numpy.array([
  [ 0.,  0.],
  [ 1.,  1.],
  [ 2.,  2.],
  [-1., -1.],
  [-2., -2.],
])

for x in data:
  print multivariate_normal.pdf(x, mean=mu, cov=Sigma)

*/

int main(int argc, char **argv) {
  MultivariateGaussian_family* mvn1 = MultivariateGaussian(1);

  gsl_vector *mu1 = gsl_vector_alloc(mvn1->dim);
  gsl_vector_set(mu1, 0, 1);

  gsl_matrix *sigma1 = gsl_matrix_alloc(mvn1->dim, mvn1->dim);
  gsl_matrix_set(sigma1, 0, 0, 3.);

  MultivariateGaussian_source* lambda1 = MultivariateGaussian_create_source(mvn1, mu1, sigma1);

  double x = 0;
  fprintf(stderr, "%f\n", mvn1->pdf(mvn1, &x, (param*)lambda1));

  Gaussian_family* g = Gaussian();
  Gaussian_source* lambdaG = Gaussian_create_source(g, 1., 3.);
  fprintf(stderr, "%f\n", g->pdf(g, &x, (param*)lambdaG));
  
  MultivariateGaussian_family* mvn2 = MultivariateGaussian(2);
  fprintf(stderr, "Family: %s\n", mvn2->name);

  /* /\* unsigned int n = 1000000; *\/ */
  /* /\* double x_min = -1.; *\/ */
  /* /\* double x_max =  1.; *\/ */
  /* /\* double dx = (x_max - x_min) / n; *\/ */
  /* /\* double y_min = -1.; *\/ */
  /* /\* double y_max =  1.; *\/ */
  /* /\* double dy = (y_max - y_min) / n; *\/ */
  /* /\* double out; *\/ */
  double in[2] = {0, 0};
 
  gsl_vector *mu2 = gsl_vector_alloc(mvn2->dim);
  gsl_vector_set(mu2, 0, 3);
  gsl_vector_set(mu2, 1, 5);

  gsl_matrix *sigma2 = gsl_matrix_alloc(mvn2->dim, mvn2->dim);
  gsl_matrix_set(sigma2, 0, 0, 1.0);
  gsl_matrix_set(sigma2, 1, 0, 0.6);
  gsl_matrix_set(sigma2, 0, 1, 0.6);
  gsl_matrix_set(sigma2, 1, 1, 2.0);

  MultivariateGaussian_source* lambda2 = MultivariateGaussian_create_source(mvn2, mu2, sigma2);

  /* /\* in[0] = x_min; *\/ */
  /* /\* for(int i=0; i<n; i++) { *\/ */
  /* /\*   in[1] = y_min; *\/ */
  /* /\*   for(int j=0; j<n; j++) { *\/ */
  /* /\*     out = mvn2->pdf(mvn2, in, (param)lambda0); *\/ */
  /* /\*     for(int d=0; d<mvn2->dim; d++) *\/ */
  /* /\*       printf("%f ", in[d]); *\/ */
  /* /\*     printf("%f\n", out); *\/ */
  /* /\*     in[1] += dy; *\/ */
  /* /\*   } *\/ */
  /* /\*   in[0] += dx; *\/ */
  /* /\* } *\/ */

  in[0] = 0;
  in[1] = 0;
  fprintf(stderr, "%f\n", mvn2->pdf(mvn2, in, (param*)lambda2));

  in[0] = 1;
  in[1] = 1;
  fprintf(stderr, "%f\n", mvn2->pdf(mvn2, in, (param*)lambda2));

  in[0] = 2;
  in[1] = 2;
  fprintf(stderr, "%f\n", mvn2->pdf(mvn2, in, (param*)lambda2));

  in[0] = -1;
  in[1] = -1;
  fprintf(stderr, "%f\n", mvn2->pdf(mvn2, in, (param*)lambda2));

  in[0] = -2;
  in[1] = -2;
  fprintf(stderr, "%f\n", mvn2->pdf(mvn2, in, (param*)lambda2));

  const gsl_rng_type *T;
  gsl_rng_env_setup();
  T = gsl_rng_default;
  gsl_rng *rng = gsl_rng_alloc(T);

  unsigned int n2 = 10000;
  double *data = malloc(n2 * mvn2->dim * sizeof(double));
  char   *mask = malloc(n2 * sizeof(char));

  for (unsigned int i = 0; i < n2; i++) {
    mask[i] = 1;
    double *r1 = mvn2->rand(mvn2, (param*)lambda2, rng);
    printf("%f %f\n", r1[0], r1[1]);
    data[mvn2->dim * i + 0] = r1[0];
    data[mvn2->dim * i + 1] = r1[1];
  }

  MultivariateGaussian_expectation* eta1 = mvn2->new_param(mvn2, EXPECTATION);
  eta1->eta1 = gsl_vector_alloc(mvn2->dim);
  eta1->eta2 = gsl_matrix_alloc(mvn2->dim, mvn2->dim);

  MultivariateGaussian_expectation* eta2 = mvn2->new_param(mvn2, EXPECTATION);
  eta2->eta1 = gsl_vector_alloc(mvn2->dim);
  eta2->eta2 = gsl_matrix_alloc(mvn2->dim, mvn2->dim);

  /* printf("ex vec addr %p %p\n", (void*)(eta1->eta1), (void*)(eta2->eta1)); */
  /* printf("ex vec size %zu %zu\n", eta1->eta1->size, eta2->eta1->size); */

  /* printf("ex mat addr %p %p\n", (void*)(eta1->eta2), (void*)(eta2->eta2)); */
  /* printf("ex mat size1 %zu %zu\n", eta1->eta2->size1, eta2->eta2->size1); */
  /* printf("ex mat size2 %zu %zu\n", eta1->eta2->size2, eta2->eta2->size2); */

  mvn2->add(mvn2, (param*)eta1, (param*)eta2, (param*)eta2);
  
  MultivariateGaussian_expectation* eta3 = mvn2->new_param(mvn2, EXPECTATION);
  eta3->eta1 = gsl_vector_calloc(mvn2->dim);
  eta3->eta2 = gsl_matrix_calloc(mvn2->dim, mvn2->dim);

  MultivariateGaussian_source* lambda5 = mvn2->new_param(mvn2, SOURCE);
  lambda5->mu    = gsl_vector_calloc(mvn2->dim);
  lambda5->Sigma = gsl_matrix_calloc(mvn2->dim, mvn2->dim);

  mvn2->mle(mvn2, n2, data, mask, (param*)eta3);
  mvn2->fprint(stderr, mvn2, (param*)eta3);
  mvn2->eta2lambda(mvn2, (param*)eta3, (param*)lambda5);
  mvn2->fprint(stderr, mvn2, (param*)lambda5);
  
  return 0;
}
